"""
Command-line interface for libratom
"""

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

PATH_METAVAR = "<path>"
INT_METAVAR = "<n>"
VERSION_METAVAR = "<version>"
MODEL_METAVAR = "<model>"
